@@ -23,10 +23,13 @@
 #include <stdio.h>
 
 struct proxy_interface;
+struct property_value_options;
 
 #define IWD_ADAPTER_INTERFACE          "net.connman.iwd.Adapter"
+#define IWD_ACCESS_POINT_INTERFACE     "net.connman.iwd.AccessPoint"
+#define IWD_AD_HOC_INTERFACE           "net.connman.iwd.AdHoc"
 #define IWD_DEVICE_INTERFACE           "net.connman.iwd.Device"
-#define IWD_KNOWN_NETWORKS_INTREFACE   "net.connman.iwd.KnownNetworks"
+#define IWD_KNOWN_NETWORK_INTREFACE    "net.connman.iwd.KnownNetwork"
 #define IWD_NETWORK_INTERFACE          "net.connman.iwd.Network"
 #define IWD_WSC_INTERFACE              "net.connman.iwd.WiFiSimpleConfiguration"
 
@@ -35,9 +38,12 @@ typedef bool (*proxy_property_match_func_t) (const void *a, const void *b);
 struct proxy_interface_property {
 	const char *name;
 	const char *type;
-	void (*set)(void *data, struct l_dbus_message_iter *variant);
+	void (*update)(void *data, struct l_dbus_message_iter *variant);
 	const char *(*tostr)(const void *data);
 	const bool is_read_write;
+	bool (*append)(struct l_dbus_message_builder *builder,
+							const char *value_str);
+	const struct property_value_options *options;
 };
 
 struct proxy_interface_type_ops {
@@ -57,6 +63,13 @@ struct proxy_interface_type {
 	const struct proxy_interface_type_ops *ops;
 };
 
+char *proxy_property_completion(
+			const struct proxy_interface_property *properties,
+			const char *text, int state);
+
+bool proxy_property_set(const struct proxy_interface *proxy, const char *name,
+			const char *value_str, l_dbus_message_func_t callback);
+
 struct proxy_interface *proxy_interface_find(const char *interface,
 							const char *path);
 
@@ -64,6 +77,9 @@ struct l_queue *proxy_interface_find_all(const char *interface,
 					proxy_property_match_func_t function,
 					const void *value);
 
+bool proxy_interface_is_same(const struct proxy_interface *a,
+					const struct proxy_interface *b);
+
 bool dbus_message_has_error(struct l_dbus_message *message);
 
 bool proxy_interface_method_call(const struct proxy_interface *proxy,
@@ -77,7 +93,7 @@ void proxy_properties_display(const struct proxy_interface *proxy,
 char *proxy_property_str_completion(const struct proxy_interface_type *type,
 					proxy_property_match_func_t function,
 					const char *property_name,
-					const char *value, int state);
+					const void *value, int state);
 
 void *proxy_interface_get_data(const struct proxy_interface *proxy);
 const char *proxy_interface_get_interface(const struct proxy_interface *proxy);
